/* jcifs smb client library in Java
 * Copyright (C) 2000  "Michael B. Allen" <jcifs at samba dot org>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package jcifs.smb;

import java.io.IOException;
import jcifs.UniAddress;

/**
 * There are hundreds of error codes that may be returned by a CIFS server. Rather than represent each with it's own <code>Exception</code> class, this class represents all of them. For many of the popular error codes, constants and text messages like "The device is not ready" are provided.
<p>
 * See <a href="../../../authhandler.html">jCIFS Exceptions and NtlmAuthenticator</a> for more information about <code>SmbException</code>.
 */

public class SmbException extends IOException {

	// error classes
	public static final int SUCCESS = 0;
	public static final int ERRDOS  = 0x01;
	public static final int ERRSRV  = 0x02;
	public static final int ERRHRD  = 0x03;
	public static final int ERRCMD  = 0xFF;
	public static final int ERRRAP  = 0x10;
	public static final int ERRCLI  = 0x20;

	// dos error codes
/**
 *  Invalid function
 */
	public static final int ERRbadfunc	  = 1;
/**
 * File not found
 */
	public static final int ERRbadfile	  = 2;
/**
 * Directory invalid
 */
	public static final int ERRbadpath	  = 3;
/**
 * Access denied
 */
	public static final int ERRnoaccess   = 5;
	public static final int ERRbadfid     = 6;
/**
 * Not enough storage is available to process this command
 */
	public static final int ERRnospace    = 8;
/**
 * The file is being used by another process
 */
	public static final int ERRbadshare   = 32;
/**
 * A duplicate name exists on the network
 */
	public static final int ERRdupname    = 52;
/**
 * The network name cannot be found
 */
	public static final int ERRbadnetname = 67;
	public static final int ERRnomoreconn = 71;
	public static final int ERRbadparm	  = 87;
/**
 * Cannot create a file when that file already exists
 */
	public static final int ERRfilexists  = 80;
/**
 * Too many Uids active on this session
 */
	public static final int ERRtoomanyuids = 90;
/**
 * The pipe has been ended
 */
	public static final int ERRbrokenpipe = 109;
/**
 * Invalid name
 */
	public static final int ERRinvname	  = 123;
/**
 * The directory is not empty
 */
	public static final int ERRnotempty   = 145;
/**
 * File exists
 */
	public static final int ERRexists     = 183;
/**
 * All pipe instances are busy
 */
	public static final int ERRpipebusy   = 231;
/**
 * The pipe is being closed
 */
	public static final int ERRnodata     = 232;
/**
 * No process is on the other end of the pipe
 */
	public static final int ERRnoproc     = 233;

	// srv error codes
/**
 * Non-specific error code
 */
	public static final int ERRerror           = 1;
/**
 * Bad password
 */
	public static final int ERRbadpw           = 2;
/**
 * Reserved (a.k.a. STATUS_PATH_NOT_COVERED)
 */
	public static final int ERRreserved  = 3;
/**
 * The client does not have the necessary access rights for the requested function
 */
	public static final int ERRaccess          = 4;
	public static final int ERRinvnid          = 5;
	public static final int ERRinvnetname      = 6;
	public static final int ERRbaduid          = 91;
/**
 * The user account has expired
 */
	public static final int ERRaccountExpired  = 2239;
/**
 * The user is not allowed to access this server from this client
 */
	public static final int ERRbadClient       = 2240;
/**
 * The user is not permitted to access the server at this time
 */
	public static final int ERRbadLogonTime    = 2241;
/**
 * The password of the user has expired
 */
	public static final int ERRpasswordExpired = 2242;

/**
 * The list of servers for this workgroup is not currently available
 */
	public static final int ERRnolisting = 6118;

	// hrd error codes
/**
 * Attempt to write on write-protected media
 */
	public static final int ERRnowrite  = 19;
/**
 * The device is not ready
 */
	public static final int ERRnotready = 21;
/**
 * General failure
 */
	public static final int ERRgeneral = 31;

	// RAP transaction status codes
	public static final int NERR_Success                = 0;
	public static final int ERROR_MORE_DATA             = 234;
	public static final int NERR_ServerNotStarted       = 2114;
	public static final int NERR_BasicTransactConfig    = 2141;

	// cli error codes (defined for jCIFS)
	public static final int ERRserverTimeout = 5000;
	public static final int ERRbadDialect    = 5001;
	public static final int ERRioe           = 5002;
	public static final int ERRlistFiles     = 5003;
	public static final int ERRunknownHost   = 5004;
	public static final int ERRinappro       = 5005;
	public static final int ERRunknownType   = 5006;
	public static final int ERRdfs           = 5007;
	public static final int ERRimpossible    = 5008;

	int errorClass;
	int errorCode;

	static String getErrorString( int errorClass, int errorCode ) {
		String result = "";
		switch( errorClass ) {
			case SUCCESS:
				result += "The operation completed successfully";
				break;
			case ERRDOS:
				switch( errorCode ) {
					case ERRbadfunc:
						result += "Invalid function";
						break;
					case ERRbadfile:
						result += "File not found";
						break;
					case ERRbadpath:
						result += "Directory invalid";
						break;
					case ERRnoaccess:
						result += "Access denied";
						break;
					case ERRbadparm:
						result += "Invalid parameter";
						break;
					case ERRinvname:
						result += "Invalid name";
						break;
					case ERRnotempty:
						result += "The directory is not empty";
						break;
					case ERRfilexists:
						result += "Cannot create a file when that file already exists";
						break;
					case ERRbadfid:
						result += "Invalid file handle";
						break;
					case ERRnospace:
						result += "Not enough storage is available to process this command";
						break;
					case ERRbadshare:
						result += "The file is being used by another process";
						break;
					case ERRdupname:
						result += "A duplicate name exists on the network";
						break;
					case ERRnotready:
						result += "The device is not ready";
						break;
					case ERRbadnetname:
						result += "The network name cannot be found";
						break;
					case ERRexists:
						result += "File exists";
						break;
					case ERRpipebusy:
						result += "All pipe instances are busy";
						break;
					case ERRnodata:
						result += "The pipe is being closed";
						break;
					case ERRnoproc:
						result += "No process is on the other end of the pipe";
						break;
					case ERRbrokenpipe:
						result += "The pipe has been ended";
						break;
					case ERRnomoreconn:
						result += "No more connections can be made to this remote computer at this time because there are already as many connections as the computer can accept";
						break;
					default:
						result += "No description available [ERRDOS/" + errorCode + "]";
				}
				break;
			case ERRSRV:
				switch( errorCode ) {
					case ERRerror:
						result += "Non-specific error code";
						break;
					case ERRbadpw:
						result += "Bad password";
						break;
					case ERRreserved:
						result += "Reserved";
						break;
					case ERRinvnid:
						result += "The Tid specified was invalid";
						break;
					case ERRinvnetname:
						result += "Invalid network name in tree connect, service not found";
						break;
					case ERRbaduid:
						result += "The UID is not known as a valid user identifier on this session";
						break;
					case ERRaccess:
						result += "The client does not have the necessary access rights for the requested function";
						break;
					case ERRaccountExpired:
						result += "The user account has expired";
						break;
					case ERRbadClient:
						result += "The user is not allowed to access this server from this client";
						break;
					case ERRbadLogonTime:
						result += "The user is not permitted to access the server at this time";
						break;
					case ERRpasswordExpired:
						result += "The password of the user has expired";
						break;
					case ERRtoomanyuids:
						result += "Too many Uids active on this session";
						break;
					default:
						result += "No description available [ERRSRV/" + errorCode + "]";
				}
				break;
			case ERRHRD:
				switch( errorCode ) {
					case ERRnowrite:
						result += "Attempt to write on write-protected media";
						break;
					case ERRnotready:
						result += "The device is not ready";
						break;
					case ERRgeneral:
						result += "General failure";
						break;
					default:
						result += "No description available [ERRHRD/" + errorCode + "]";
				}
				break;
			case ERRCMD:
				result += "Command was not in the \"SMB\" format";
				break;
			case ERRRAP:
				// I don't think this should be here.
				// Should switch to NT status codes.
				switch( errorCode ) {
					case NERR_Success:
						result += "The operation completed successfully";
						break;
					case ERRnowrite:
						NERR_ServerNotStarted:
						result += "The Server service is not started";
						break;
					case NERR_BasicTransactConfig:
						result += "The server is not configured for transactions";
						break;
					case ERRnolisting:
						result += "The list of servers for this workgroup is not currently available";
						break;
					default:
						result += "No description available [ERRRAP/" + errorCode + "]";
				}
				break;
			case ERRCLI:
				switch( errorCode ) {
					case ERRserverTimeout:
						result += "Timeout waiting for response from server.";
						break;
					case ERRbadDialect:
						result += "Failed to negotiate dialect with server.";
						break;
					case ERRioe:
						result += "An IO error occured.";
						break;
					case ERRlistFiles:
						result += "The derived URL was malformed or a name service error occured during a listFiles() operation (this should not be possible, please report it to jcifs at samba dot org)";
						break;
					case ERRunknownHost:
						result += "The hostname retrived from the Browser service is no longer valid";
						break;
					case ERRinappro:
						result += "Inappropriate operation";
						break;
					case ERRunknownType:
						result += "Unknown resource type";
						break;
					case ERRdfs:
						result += "Invalid DFS operation";
						break;
					default:
						result += "No description available [ERRCLI/" + errorCode + "]";
				}
				break;
			default:
				result += "unknown error class: " + errorClass;
		}
		return result;
	}
	static String getErrorString( int code ) {
		if( code == 0 ) {
			return "0x00000000";
		}
		return getErrorString( code & 0xFF, ( code >> 16 ) & 0xFFFF );
	}

	SmbException() {
	}
	SmbException( int code ) {
		super( getErrorString( code & 0xFF, ( code >> 16 ) & 0xFFFF ));
		errorClass = code & 0xFF;
		errorCode = ( code >> 16 ) & 0xFFFF;
	}
	SmbException( int errorClass, int errorCode ) {
		super( getErrorString( errorClass, errorCode ));
		this.errorClass = errorClass;
		this.errorCode = errorCode;
	}
	SmbException( int errorClass, int errorCode, UniAddress addr ) {
		super( getErrorString( errorClass, errorCode ) + " " + addr.toString() );
		this.errorClass = errorClass;
		this.errorCode = errorCode;
	}
	SmbException( int errorClass, int errorCode, String message ) {
		super( message );
		this.errorClass = errorClass;
		this.errorCode = errorCode;
	}
	SmbException( int code, String message ) {
		super( message );
		errorClass = code & 0xFF;
		errorCode = ( code >> 16 ) & 0xFFFF;
	}

	/**
	 * Retrieve the error class associated with this error.
	 */
	public int getErrorClass() {
		return errorClass;
	}

	/**
	 * Retrieve the error code associated with this error.
	 */
	public int getErrorCode() {
		return errorCode;
	}
}
